class Stream:
    """
    A fluid stream carrying heat and consisting of one or multiple segments.
    """

    def __init__(self, *segments):
        """
        `segments` must meet the following conditions:
        * It must not be an empty sequence.
        * It must be continuous, i.e. there must be no temperature gaps between
          adjacent segments.
        If one of those conditions is not met, a ValueError is raised.
        """
        Stream._check_segments(segments)
        self._segments = segments

    @property
    def heat_flow(self):
        return sum(s.heat_flow for s in self.segments)

    @property
    def supply_temp(self):
        return self.segments[0].supply_temp

    @property
    def target_temp(self):
        return self.segments[-1].target_temp

    @property
    def segments(self):
        """
        Returns a list of the stream's segments.
        """
        return list(self._segments)

    @property
    def neutral_segments(self):
        """
        Returns a list of the stream's neutral segments.
        """
        return [s for s in self.segments if s.heat_flow == 0]

    @property
    def cold_segments(self):
        """
        Returns a list of the stream's cold segments.
        """
        return [s for s in self.segments if s.heat_flow < 0]

    @property
    def hot_segments(self):
        """
        Returns a list of the stream's hot segments.
        """
        return [s for s in self.segments if s.heat_flow > 0]

    @staticmethod
    def _check_segments(segments):
        if not segments:
            raise ValueError("`segments` empty")

        prev = None
        for current in segments:
            if prev is not None and current.supply_temp != prev.target_temp:
                raise ValueError(
                    "Temperature gap between adjacent segments: {}, {}".format(
                        prev, current
                    )
                )

            prev = current

    def __eq__(self, other):
        return self.segments == other.segments

    def __repr__(self):
        return "{}{}".format(type(self).__qualname__, self._segments)


def make_stream(heat_flow, supply_temp, target_temp, temp_shift=None):
    """
    Convenience function to create a single segment stream.
    """
    return make_segmented_stream([heat_flow, supply_temp, target_temp, temp_shift])


def make_segmented_stream(*segments):
    """
    Convenience function to create a segmented stream. Each argument represents
    one segment. It must have the following form:
    [heat_flow, supply_temp, target_temp, temp_shift=None]
    """
    from pina.segments import make_segment

    return Stream(*(make_segment(*s) for s in segments))



from matplotlib import pyplot as plt

from pina import PinchAnalyzer, make_segmented_stream, make_stream

# Some customizations to make the plots pretty
HCC_STYLE = {
    "color": "tab:red",
    "linestyle": "--",
    "label": "HCC",
}
CCC_STYLE = {
    "color": "tab:blue",
    "linestyle": "-",
    "label": "CCC",
}
GCC_STYLE = {
    "color": "k",
    "linestyle": "-",
    "label": "GCC",
}
PLOT_PARAMS = {
    "title": "",
    "hcc_ccc_title": "Hot and cold composite curves",
    "hcc_ccc_xlabel": "Heat flow [kW]",
    "hcc_ccc_ylabel": "Actual temperature [\u2103]",
    "gcc_title": "Grand composite curve",
    "gcc_xlabel": "Net heat flow [kW]",
    "gcc_ylabel": "Shifted temperature [\u2103]",
    "figsize": (16, 7),
    "hcc_style": HCC_STYLE,
    "ccc_style": CCC_STYLE,
    "gcc_style": GCC_STYLE,
}


def describe(pinch_analyzer):
    """
    Returns a string with the energy targets and the pinch temperature(s).
    """
    names = [
        "Heating demand:",
        "Cooling demand:",
        "Hot utility target:",
        "Cold utility target:",
        "Heat recovery target:",
        "Pinch temperature(s):",
    ]
    values = [
        pinch_analyzer.heating_demand,
        pinch_analyzer.cooling_demand,
        pinch_analyzer.hot_utility_target,
        pinch_analyzer.cold_utility_target,
        pinch_analyzer.heat_recovery_target,
        pinch_analyzer.pinch_temps,
    ]
    max_width = max(len(n) for n in names)
    return "".join(
        "{:{width}} {}\n".format(n, v, width=max_width) for n, v in zip(names, values)
    )


def plot_results(pinch_analyzer, param_dict):
    """
    Creates two plots: One of the hot and cold composite curves and one of the
    grand composite curve. Returns a tuple with the figure and axes objects.
    """
    fig, ax = plt.subplots(1, 2, figsize=param_dict["figsize"])
    fig.suptitle(param_dict["title"], fontweight="bold")

    ax[0].plot(*pinch_analyzer.hot_composite_curve, **param_dict["hcc_style"])
    ax[0].plot(*pinch_analyzer.cold_composite_curve, **param_dict["ccc_style"])
    ax[0].legend()
    ax[0].set_title(param_dict["hcc_ccc_title"])
    ax[0].set_xlabel(param_dict["hcc_ccc_xlabel"])
    ax[0].set_ylabel(param_dict["hcc_ccc_ylabel"])

    ax[1].plot(*pinch_analyzer.grand_composite_curve, **param_dict["gcc_style"])
    ax[1].legend()
    ax[1].set_title(param_dict["gcc_title"])
    ax[1].set_xlabel(param_dict["gcc_xlabel"])
    ax[1].set_ylabel(param_dict["gcc_ylabel"])

    # Make the y-axis equal in both plots
    ylims = (*ax[0].get_ylim(), *ax[1].get_ylim())
    minmax_ylims = (min(ylims), max(ylims))
    ax[0].set_ylim(minmax_ylims)
    ax[1].set_ylim(minmax_ylims)

    return fig, ax


def four_stream():
    # Four stream example from
    # Pinch Analysis and Process Integration:
    # A User Guide On Process Integration for the Efficient Use of Energy,
    # Second edition, Ian C. Kemp, page 4
    param_dict = dict(PLOT_PARAMS)
    param_dict["title"] = "Four stream example"

    min_temp_diff = 10
    default_temp_shift = min_temp_diff / 2

    streams = [
        make_stream(-230, 20, 135),
        make_stream(330, 170, 60),
        make_stream(-240, 80, 140),
        make_stream(180, 150, 30),
    ]

    analyzer = PinchAnalyzer(default_temp_shift)
    analyzer.add_streams(*streams)

    print(param_dict["title"])
    print(describe(analyzer))

    fig, ax = plot_results(analyzer, param_dict)
    plt.show()


def aromatics_plant():
    # Aromatics plant from
    # Pinch Analysis and Process Integration:
    # A User Guide On Process Integration for the Efficient Use of Energy,
    # Second edition, Ian C. Kemp, page 330
    param_dict = dict(PLOT_PARAMS)
    param_dict.update(
        {
            "title": "Aromatics plant",
            "hcc_ccc_xlabel": "Heat flow [ttc/h]",
            "gcc_xlabel": "Net heat flow [ttc/h]",
        }
    )

    min_temp_diff = 10
    default_temp_shift = min_temp_diff / 2

    # Some of the streams consist of multiple segments with individual heat
    # capacities
    streams = [
        make_segmented_stream([-13.9, 102, 229], [-8.3, 229, 327]),
        make_segmented_stream([13.9, 327, 174], [9, 174, 92], [4.2, 92, 50]),
        make_stream(-9, 35, 164),
        make_segmented_stream([-7.2, 140, 176], [-25.2, 176, 367], [-16.4, 367, 500]),
        make_stream(25.2, 495, 307),
        make_segmented_stream(
            [7.2, 220, 160], [3.3, 160, 144], [4.1, 144, 125], [11.6, 125, 59]
        ),
        make_stream(-3.3, 80, 123),
        make_stream(-6.8, 59, 169),
        make_segmented_stream([6.8, 220, 130], [3.8, 130, 67]),
        make_stream(-4.1, 85, 125),
        make_stream(-32.5, 480, 500),
    ]

    analyzer = PinchAnalyzer(default_temp_shift)
    analyzer.add_streams(*streams)

    print(param_dict["title"])
    print(describe(analyzer))

    plot_results(analyzer, param_dict)
    plt.show()


def evaporator_dryer_plant():
    # Evaporator/dryer plant from
    # Pinch Analysis and Process Integration:
    # A User Guide On Process Integration for the Efficient Use of Energy,
    # Second edition, Ian C. Kemp, page 351
    param_dict = dict(PLOT_PARAMS)
    param_dict["title"] = "Evaporator/dryer plant"

    min_temp_diff = 5.5
    default_temp_shift = min_temp_diff / 2

    # Some of the streams have equal supply and target temperatures, meaning
    # that they transfer latent heat
    evaporator_streams = [
        make_stream(-183, 10, 70),
        make_stream(-198, 37.8, 87.8),
        make_stream(-1005.5, 79.4, 79.4),
        make_stream(1039, 79.4, 79.4),
        make_stream(232, 86.9, 10),
        make_stream(-643, 48.8, 48.8),
        make_stream(714, 43.3, 43.3),
        make_segmented_stream([-6.5, 48.8, 54.4], [-263.5, 54.4, 93.3]),
        make_stream(260, 43.3, 43.3),
        make_stream(57, 43.3, 10),
    ]

    dryer_streams = [
        make_stream(-93.5, 55, 55),
        make_stream(-254, 41, 41),
        make_stream(-124, 60, 60),
        # Some optional streams
        # make_stream(149, 41, 13),
        # make_stream(140, 60, 13),
        # make_stream(189, 55, 13),
    ]

    analyzer = PinchAnalyzer(default_temp_shift)
    analyzer.add_streams(*evaporator_streams)
    analyzer.add_streams(*dryer_streams)

    print(param_dict["title"])
    print(describe(analyzer))

    fig, ax = plot_results(analyzer, param_dict)
    plt.show()


def multiple_pinches():
    # A made-up example with multiple pinches
    param_dict = dict(PLOT_PARAMS)
    param_dict["title"] = "Multiple pinches"

    min_temp_diff = 10
    default_temp_shift = min_temp_diff / 2

    cold_1 = make_stream(-360, 30, 150)
    hot_1 = make_segmented_stream(
        [20, 140, 100],
        [90, 100, 100],  # A latent segment to model condensation
        [30, 100, 60],
    )
    # The PinchAnalyzer's default temp_shift will be ignored for this stream.
    cold_2 = make_stream(-20, 5, 20, temp_shift=2)
    hot_2 = make_stream(160, 60, 20)

    analyzer = PinchAnalyzer(default_temp_shift)
    analyzer.add_streams(cold_1, hot_1, cold_2, hot_2)

    print(param_dict["title"])
    print(describe(analyzer))

    fig, ax = plot_results(analyzer, param_dict)
    plt.show()


def banana():
    param_dict = dict(PLOT_PARAMS)
    param_dict["title"] = "Banana"

    min_temp_diff = 5
    default_temp_shift = min_temp_diff / 2

    cold = make_segmented_stream(
        [-4, 10, 10],
        [-4, 10, 10.25],
        [-3, 10.25, 10.5],
        [-4, 10.5, 11],
        [-5, 11, 12],
        [-10, 12, 15],
        [-10, 15, 18],
        [-5, 18, 20],
        [-5, 20, 22],
        [-4.5, 22, 24],
        [-6, 24, 28],
        [-5.5, 28, 34],
        [-3, 34, 39],
        [-4, 39, 48],
        [-0.5, 48, 49.5],
        [-0.5, 49.5, 52],
        [-0.5, 52, 55],
        [0, 55, 61],
        [-0.5, 61, 65.5],
        [-0.5, 65.5, 67],
        [-2, 67, 70],
        [-1, 70, 71],
        [-1.5, 71, 72],
    )

    hot = make_segmented_stream(
        [1, 74, 71],
        [1, 71, 68.5],
        [1, 68.5, 67],
        [1, 67, 65.5],
        [1, 65.5, 64.5],
        [1, 64.5, 63.5],
        [2.5, 63.5, 63],
        [2.5, 63, 62],
        [2.5, 62, 60],
        [1.5, 60, 58.5],
        [1, 58.5, 57],
        [5, 57, 50],
        [5, 50, 45],
        [5, 45, 41],
        [5, 41, 38],
        [5, 38, 36],
        [5, 36, 34],
        [5, 34, 32],
        [5, 32, 30],
        [5, 30, 28],
        [6, 28, 25],
        [5, 25, 21],
        [4, 21, 17],
    )

    analyzer = PinchAnalyzer(default_temp_shift)
    analyzer.add_streams(cold, hot)

    print(param_dict["title"])
    print(describe(analyzer))

    fig, ax = plot_results(analyzer, param_dict)
    plt.show()


if __name__ == "__main__":
    four_stream()
    aromatics_plant()
    evaporator_dryer_plant()
    multiple_pinches()
    banana()



from pina.heat_cascade import HeatCascade


class PinchAnalyzer:
    """
    This class performs pinch analysis calculations on a list of streams:
    heating and cooling requirements, heat cascades and the pinch
    temperature(s).
    """

    def __init__(self, default_temp_shift=None):
        self._default_temp_shift = default_temp_shift
        self._streams = []
        self._pinch_temps = []
        self._cold_utility_target = 0
        self._hot_utility_target = 0

        self._cold_cascade = HeatCascade()
        self._hot_cascade = HeatCascade()
        self._shifted_cold_cascade = HeatCascade()
        self._shifted_hot_cascade = HeatCascade()
        self._grand_cascade = HeatCascade()

    @property
    def cooling_demand(self):
        return self._hot_cascade.net_heat_flow()

    @property
    def heating_demand(self):
        return self._cold_cascade.net_heat_flow()

    @property
    def cold_utility_target(self):
        return self._cold_utility_target

    @property
    def hot_utility_target(self):
        return self._hot_utility_target

    @property
    def heat_recovery_target(self):
        return self.heating_demand - self.hot_utility_target

    @property
    def pinch_temps(self):
        return self._pinch_temps

    @property
    def default_temp_shift(self):
        return self._default_temp_shift

    @property
    def streams(self):
        return self._streams

    @property
    def cold_composite_curve(self):
        return self._cold_cascade.cumulative_heat_flow(self.cold_utility_target)

    @property
    def hot_composite_curve(self):
        return self._hot_cascade.cumulative_heat_flow()

    @property
    def shifted_cold_composite_curve(self):
        return self._shifted_cold_cascade.cumulative_heat_flow(self.cold_utility_target)

    @property
    def shifted_hot_composite_curve(self):
        return self._shifted_hot_cascade.cumulative_heat_flow()

    @property
    def grand_composite_curve(self):
        return self._grand_cascade.cumulative_heat_flow(self.cold_utility_target)

    def add_streams(self, *streams):
        """
        Adds the given streams to the stream group.
        """
        for s in streams:
            self._add_one(s)

        self._compute_targets()

    def _add_one(self, stream):
        self._streams.append(stream)

        for s in stream.cold_segments:
            self._cold_cascade.add_segments(s.with_absolute_heat_flow())
            shifted = s.shift(self.default_temp_shift)
            self._shifted_cold_cascade.add_segments(shifted.with_absolute_heat_flow())
            self._grand_cascade.add_segments(shifted.with_inverted_heat_flow())

        for s in stream.hot_segments:
            self._hot_cascade.add_segments(s.with_absolute_heat_flow())
            shifted = s.shift(self.default_temp_shift)
            self._shifted_hot_cascade.add_segments(shifted.with_absolute_heat_flow())
            self._grand_cascade.add_segments(shifted.with_inverted_heat_flow())

    def _compute_targets(self):
        heat_flows, temps = self._grand_cascade.cumulative_heat_flow()
        if heat_flows:
            min_heat_flow = min(heat_flows)
            self._pinch_temps = [
                t for t, h in zip(temps, heat_flows) if h == min_heat_flow
            ]
            self._cold_utility_target = heat_flows[0] - min_heat_flow
            self._hot_utility_target = heat_flows[-1] - min_heat_flow
        else:
            self._pinch_temps = []
            self._cold_utility_target = 0
            self._hot_utility_target = 0



import unittest

from pina.pinch_analyzer import PinchAnalyzer
from pina.stream import make_segmented_stream, make_stream


class TestPinchAnalyzer(unittest.TestCase):
    """
    Test class for PinchAnalyzer
    """

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_empty(self):
        analyzer = PinchAnalyzer()
        self.assertEqual(analyzer.cooling_demand, 0)
        self.assertEqual(analyzer.heating_demand, 0)
        self.assertEqual(analyzer.cold_utility_target, 0)
        self.assertEqual(analyzer.hot_utility_target, 0)
        self.assertEqual(analyzer.heat_recovery_target, 0)
        self.assertEqual(analyzer.pinch_temps, [])
        self.assertEqual(analyzer.default_temp_shift, None)
        self.assertEqual(analyzer.streams, [])
        self.assertEqual(analyzer.cold_composite_curve, ([], []))
        self.assertEqual(analyzer.hot_composite_curve, ([], []))
        self.assertEqual(analyzer.shifted_cold_composite_curve, ([], []))
        self.assertEqual(analyzer.shifted_hot_composite_curve, ([], []))
        self.assertEqual(analyzer.grand_composite_curve, ([], []))

    def test_2_stream_example(self):
        cold_stream = make_stream(-180, 20, 200)
        hot_stream = make_stream(180, 150, 50)
        analyzer = PinchAnalyzer(default_temp_shift=10)
        analyzer.add_streams(cold_stream, hot_stream)

        self.assertEqual(analyzer.cooling_demand, 180)
        self.assertEqual(analyzer.heating_demand, 180)
        self.assertEqual(analyzer.cold_utility_target, 70)
        self.assertEqual(analyzer.hot_utility_target, 70)
        self.assertEqual(analyzer.heat_recovery_target, 110)
        self.assertEqual(analyzer.pinch_temps, [140])
        self.assertEqual(analyzer.default_temp_shift, 10)

        self.assertEqual(analyzer.streams, [cold_stream, hot_stream])

        expected_cold_composite_curve = ([70, 250], [20, 200])
        self.assertEqual(analyzer.cold_composite_curve, expected_cold_composite_curve)

        expected_hot_composite_curve = ([0, 180], [50, 150])
        self.assertEqual(analyzer.hot_composite_curve, expected_hot_composite_curve)

        expected_shifted_cold_composite_curve = ([70, 250], [30, 210])
        self.assertEqual(
            analyzer.shifted_cold_composite_curve, expected_shifted_cold_composite_curve
        )

        expected_shifted_hot_composite_curve = ([0, 180], [40, 140])
        self.assertEqual(
            analyzer.shifted_hot_composite_curve, expected_shifted_hot_composite_curve
        )

        expected_grand_composite_curve = ([70, 80, 0, 70], [30, 40, 140, 210])
        self.assertEqual(analyzer.grand_composite_curve, expected_grand_composite_curve)

    def test_extended_pinch(self):
        cold_stream = make_stream(-225, 20, 95)
        hot_stream = make_stream(267, 95, 6)
        analyzer = PinchAnalyzer(default_temp_shift=1)
        analyzer.add_streams(cold_stream, hot_stream)

        self.assertEqual(analyzer.cooling_demand, 267)
        self.assertEqual(analyzer.heating_demand, 225)
        self.assertEqual(analyzer.cold_utility_target, 48)
        self.assertEqual(analyzer.hot_utility_target, 6)
        self.assertEqual(analyzer.heat_recovery_target, 219)
        self.assertEqual(analyzer.pinch_temps, [21, 94])
        self.assertEqual(analyzer.default_temp_shift, 1)

    def test_2_pinches(self):
        cold_stream = make_segmented_stream(
            [-50, 50, 100], [-40, 100, 100], [-100, 100, 200]
        )
        hot_stream = make_stream(260, 150, 20)
        analyzer = PinchAnalyzer(default_temp_shift=5)
        analyzer.add_streams(cold_stream, hot_stream)

        self.assertEqual(analyzer.cooling_demand, 260)
        self.assertEqual(analyzer.heating_demand, 190)
        self.assertEqual(analyzer.cold_utility_target, 130)
        self.assertEqual(analyzer.hot_utility_target, 60)
        self.assertEqual(analyzer.heat_recovery_target, 130)
        self.assertEqual(analyzer.pinch_temps, [105, 145])
        self.assertEqual(analyzer.default_temp_shift, 5)

    def test_cold_stream_only(self):
        cold_stream = make_stream(-40, 100, 100)
        analyzer = PinchAnalyzer(default_temp_shift=5)
        analyzer.add_streams(cold_stream)

        self.assertEqual(analyzer.cooling_demand, 0)
        self.assertEqual(analyzer.heating_demand, 40)
        self.assertEqual(analyzer.cold_utility_target, 0)
        self.assertEqual(analyzer.hot_utility_target, 40)
        self.assertEqual(analyzer.heat_recovery_target, 0)
        self.assertEqual(analyzer.pinch_temps, [105])
        self.assertEqual(analyzer.default_temp_shift, 5)

    def test_hot_stream_only(self):
        hot_stream = make_stream(260, 150, 20)
        analyzer = PinchAnalyzer(default_temp_shift=5)
        analyzer.add_streams(hot_stream)

        self.assertEqual(analyzer.cooling_demand, 260)
        self.assertEqual(analyzer.heating_demand, 0)
        self.assertEqual(analyzer.cold_utility_target, 260)
        self.assertEqual(analyzer.hot_utility_target, 0)
        self.assertEqual(analyzer.heat_recovery_target, 0)
        self.assertEqual(analyzer.pinch_temps, [145])
        self.assertEqual(analyzer.default_temp_shift, 5)

    def test_individual_temp_shift(self):
        cold_stream = make_stream(-40, 100, 100, 10)
        analyzer = PinchAnalyzer(default_temp_shift=5)
        analyzer.add_streams(cold_stream)

        self.assertEqual(analyzer.cooling_demand, 0)
        self.assertEqual(analyzer.heating_demand, 40)
        self.assertEqual(analyzer.cold_utility_target, 0)
        self.assertEqual(analyzer.hot_utility_target, 40)
        self.assertEqual(analyzer.heat_recovery_target, 0)
        self.assertEqual(analyzer.pinch_temps, [110])
        self.assertEqual(analyzer.default_temp_shift, 5)

    def test_neutralizing_streams(self):
        cold_stream = make_stream(-40, 100, 100, 0)
        hot_stream = make_stream(40, 100, 100, 0)
        analyzer = PinchAnalyzer(default_temp_shift=5)
        analyzer.add_streams(cold_stream)
        analyzer.add_streams(hot_stream)

        self.assertEqual(analyzer.cooling_demand, 40)
        self.assertEqual(analyzer.heating_demand, 40)
        self.assertEqual(analyzer.cold_utility_target, 0)
        self.assertEqual(analyzer.hot_utility_target, 0)
        self.assertEqual(analyzer.heat_recovery_target, 40)
        self.assertEqual(analyzer.pinch_temps, [])
        self.assertEqual(analyzer.default_temp_shift, 5)
        self.assertEqual(analyzer.streams, [cold_stream, hot_stream])


if __name__ == "__main__":
    unittest.main()



import unittest

from pina.heat_cascade import HeatCascade
from pina.segments import make_segment


class TestHeatCascade(unittest.TestCase):
    """
    Test class for HeatCascade
    """

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_empty(self):
        cascade = HeatCascade()
        self.assertEqual(cascade.intervals, [])
        self.assertEqual(cascade.net_heat_flow(), 0)
        self.assertEqual(cascade.cumulative_heat_flow(), ([], []))
        self.assertEqual(cascade.cumulative_heat_flow(-100), ([], []))

    def test_neutral_sensible_segment(self):
        neutral_segment = make_segment(0, 80, 120)
        cascade = HeatCascade()
        cascade.add_segments(neutral_segment)
        self.assertEqual(cascade.intervals, [])
        self.assertEqual(cascade.net_heat_flow(), 0)
        self.assertEqual(cascade.cumulative_heat_flow(), ([], []))
        self.assertEqual(cascade.cumulative_heat_flow(-100), ([], []))

    def test_cold_sensible_segment(self):
        cold_segment = make_segment(-180, 20, 200)
        cascade = HeatCascade()
        cascade.add_segments(cold_segment)
        self.assertEqual(cascade.intervals, [cold_segment])
        self.assertEqual(cascade.net_heat_flow(), -180)
        self.assertEqual(cascade.cumulative_heat_flow(), ([0, -180], [20, 200]))
        self.assertEqual(cascade.cumulative_heat_flow(50), ([50, -130], [20, 200]))

    def test_hot_sensible_segment(self):
        hot_segment = make_segment(180, 150, 50)
        cascade = HeatCascade()
        cascade.add_segments(hot_segment)
        self.assertEqual(cascade.intervals, [make_segment(180, 50, 150)])
        self.assertEqual(cascade.net_heat_flow(), 180)
        self.assertEqual(cascade.cumulative_heat_flow(), ([0, 180], [50, 150]))
        self.assertEqual(cascade.net_heat_flow(), 180)
        self.assertEqual(cascade.cumulative_heat_flow(-180), ([-180, 0], [50, 150]))

    def test_neutral_latent_segment(self):
        neutral_segment = make_segment(0, 80, 80)
        cascade = HeatCascade()
        cascade.add_segments(neutral_segment)
        self.assertEqual(cascade.intervals, [])
        self.assertEqual(cascade.net_heat_flow(), 0)
        self.assertEqual(cascade.cumulative_heat_flow(), ([], []))
        self.assertEqual(cascade.cumulative_heat_flow(-100), ([], []))

    def test_cold_latent_segment(self):
        cold_segment = make_segment(-200, 100, 100)
        cascade = HeatCascade()
        cascade.add_segments(cold_segment)
        self.assertEqual(cascade.intervals, [cold_segment])
        self.assertEqual(cascade.net_heat_flow(), -200)
        self.assertEqual(cascade.cumulative_heat_flow(), ([0, -200], [100, 100]))
        self.assertEqual(cascade.cumulative_heat_flow(-50), ([-50, -250], [100, 100]))

    def test_hot_latent_segment(self):
        hot_segment = make_segment(300, 150, 150)
        cascade = HeatCascade()
        cascade.add_segments(hot_segment)
        self.assertEqual(cascade.intervals, [hot_segment])
        self.assertEqual(cascade.net_heat_flow(), 300)
        self.assertEqual(cascade.cumulative_heat_flow(), ([0, 300], [150, 150]))
        self.assertEqual(cascade.cumulative_heat_flow(100), ([100, 400], [150, 150]))

    def test_touching_segments(self):
        cold_segments = [make_segment(-60, 20, 80), make_segment(-80, 80, 120)]
        cold_cascade = HeatCascade()
        cold_cascade.add_segments(*cold_segments)
        self.assertEqual(cold_cascade.intervals, cold_segments)
        self.assertEqual(cold_cascade.net_heat_flow(), -140)
        self.assertEqual(
            cold_cascade.cumulative_heat_flow(), ([0, -60, -140], [20, 80, 120])
        )

        hot_segments = [make_segment(80, 120, 80), make_segment(60, 80, 20)]
        hot_cascade = HeatCascade()
        hot_cascade.add_segments(*hot_segments)
        self.assertEqual(
            hot_cascade.intervals, [make_segment(60, 20, 80), make_segment(80, 80, 120)]
        )
        self.assertEqual(hot_cascade.net_heat_flow(), 140)
        self.assertEqual(
            hot_cascade.cumulative_heat_flow(), ([0, 60, 140], [20, 80, 120])
        )

    def test_detached_segments(self):
        detached_segments = [make_segment(-30, 20, 50), make_segment(-80, 80, 120)]
        cascade = HeatCascade()
        cascade.add_segments(*detached_segments)
        self.assertEqual(cascade.intervals, detached_segments)
        self.assertEqual(cascade.net_heat_flow(), -110)
        self.assertEqual(
            cascade.cumulative_heat_flow(), ([0, -30, -30, -110], [20, 50, 80, 120])
        )

    def test_overlapping_cold_segments(self):
        cold_segments = [make_segment(-65, 20, 85), make_segment(-120, 60, 120)]
        expected_intervals = [
            make_segment(-40, 20, 60),
            make_segment(-75, 60, 85),
            make_segment(-70, 85, 120),
        ]
        cascade = HeatCascade()
        cascade.add_segments(*cold_segments)
        self.assertEqual(cascade.intervals, expected_intervals)
        self.assertEqual(cascade.net_heat_flow(), -185)
        self.assertEqual(
            cascade.cumulative_heat_flow(), ([0, -40, -115, -185], [20, 60, 85, 120])
        )

    def test_overlapping_hot_segments(self):
        hot_segments = [make_segment(60, 120, 60), make_segment(130, 85, 20)]
        expected_intervals = [
            make_segment(80, 20, 60),
            make_segment(75, 60, 85),
            make_segment(35, 85, 120),
        ]
        cascade = HeatCascade()
        cascade.add_segments(*hot_segments)
        self.assertEqual(cascade.intervals, expected_intervals)
        self.assertEqual(cascade.net_heat_flow(), 190)
        self.assertEqual(
            cascade.cumulative_heat_flow(), ([0, 80, 155, 190], [20, 60, 85, 120])
        )

    def test_link_cold_sensible_segments(self):
        cold_segments = [make_segment(-120, 20, 80), make_segment(-80, 80, 120)]
        cascade = HeatCascade()
        cascade.add_segments(*cold_segments)
        self.assertEqual(cascade.intervals, [make_segment(-200, 20, 120)])
        self.assertEqual(cascade.net_heat_flow(), -200)
        self.assertEqual(cascade.cumulative_heat_flow(), ([0, -200], [20, 120]))

    def test_link_hot_sensible_segments(self):
        hot_segments = [make_segment(80, 120, 80), make_segment(120, 80, 20)]
        cascade = HeatCascade()
        cascade.add_segments(*hot_segments)
        self.assertEqual(cascade.intervals, [make_segment(200, 20, 120)])
        self.assertEqual(cascade.net_heat_flow(), 200)
        self.assertEqual(cascade.cumulative_heat_flow(), ([0, 200], [20, 120]))

    def test_link_latent_segments(self):
        latent_segments = [
            make_segment(-200, 100, 100),
            make_segment(-150, 100, 100),
            make_segment(250, 100, 100),
        ]
        expected_intervals = [make_segment(-100, 100, 100)]
        cascade = HeatCascade()
        cascade.add_segments(*latent_segments)
        self.assertEqual(cascade.intervals, expected_intervals)
        self.assertEqual(cascade.net_heat_flow(), -100)
        self.assertEqual(cascade.cumulative_heat_flow(), ([0, -100], [100, 100]))

    def test_detached_latent_segments(self):
        latent_segments = [make_segment(-200, 100, 100), make_segment(-150, 200, 200)]
        cascade = HeatCascade()
        cascade.add_segments(*latent_segments)
        self.assertEqual(cascade.intervals, latent_segments)
        self.assertEqual(cascade.net_heat_flow(), -350)
        self.assertEqual(
            cascade.cumulative_heat_flow(),
            ([0, -200, -200, -350], [100, 100, 200, 200]),
        )

    def test_add_sensible_then_latent_segment(self):
        sensible_then_latent = HeatCascade()
        sensible_then_latent.add_segments(
            make_segment(-80, 80, 120), make_segment(-200, 100, 100)
        )
        expected_intervals = [
            make_segment(-40, 80, 100),
            make_segment(-200, 100, 100),
            make_segment(-40, 100, 120),
        ]
        self.assertEqual(sensible_then_latent.intervals, expected_intervals)
        self.assertEqual(sensible_then_latent.net_heat_flow(), -280)
        self.assertEqual(
            sensible_then_latent.cumulative_heat_flow(),
            ([0, -40, -240, -280], [80, 100, 100, 120]),
        )

    def test_add_latent_then_sensible_segment(self):
        latent_then_sensible = HeatCascade()
        latent_then_sensible.add_segments(
            make_segment(-200, 100, 100), make_segment(-80, 80, 120)
        )
        expected_intervals = [
            make_segment(-40, 80, 100),
            make_segment(-200, 100, 100),
            make_segment(-40, 100, 120),
        ]
        self.assertEqual(latent_then_sensible.intervals, expected_intervals)
        self.assertEqual(latent_then_sensible.net_heat_flow(), -280)
        self.assertEqual(
            latent_then_sensible.cumulative_heat_flow(),
            ([0, -40, -240, -280], [80, 100, 100, 120]),
        )

    def test_mixed_sensible_segments(self):
        mixed_segments = [
            make_segment(-60, 60, 120),
            make_segment(-10, 60, 70),
            make_segment(130, 85, 20),
        ]
        expected_intervals = [
            make_segment(80, 20, 60),
            make_segment(15, 70, 85),
            make_segment(-35, 85, 120),
        ]
        cascade = HeatCascade()
        cascade.add_segments(*mixed_segments)
        self.assertEqual(cascade.intervals, expected_intervals)
        self.assertEqual(cascade.net_heat_flow(), 60)
        self.assertEqual(
            cascade.cumulative_heat_flow(), ([0, 80, 80, 95, 60], [20, 60, 70, 85, 120])
        )

    def test_neutralized_heat_flow(self):
        neutralizing_segments = [make_segment(-60, 60, 120), make_segment(60, 120, 60)]
        cascade = HeatCascade()
        cascade.add_segments(*neutralizing_segments)
        self.assertEqual(cascade.intervals, [])
        self.assertEqual(cascade.net_heat_flow(), 0)
        self.assertEqual(cascade.cumulative_heat_flow(), ([], []))

    def test_add_cascades(self):
        cold_cascade = HeatCascade()
        cold_cascade.add_segments(make_segment(-230, 25, 140))
        cold_cascade.add_segments(make_segment(-240, 85, 145))

        hot_cascade = HeatCascade()
        hot_cascade.add_segments(make_segment(330, 165, 55))
        hot_cascade.add_segments(make_segment(180, 145, 25))

        mixed_cascade = HeatCascade()
        mixed_cascade.add_segments(*cold_cascade.intervals)
        mixed_cascade.add_segments(*hot_cascade.intervals)
        self.assertEqual(
            mixed_cascade.intervals,
            [
                make_segment(-15, 25, 55),
                make_segment(75, 55, 85),
                make_segment(-82.5, 85, 140),
                make_segment(2.5, 140, 145),
                make_segment(60, 145, 165),
            ],
        )
        self.assertEqual(mixed_cascade.net_heat_flow(), 40)
        self.assertEqual(
            mixed_cascade.cumulative_heat_flow(),
            ([0, -15, 60, -22.5, -20, 40], [25, 55, 85, 140, 145, 165]),
        )
        self.assertEqual(
            mixed_cascade.cumulative_heat_flow(22.5),
            ([22.5, 7.5, 82.5, 0, 2.5, 62.5], [25, 55, 85, 140, 145, 165]),
        )

    def test_equality_comparison(self):
        cascade = HeatCascade()
        self.assertEqual(cascade, HeatCascade())
        compare_to = HeatCascade()
        compare_to.add_segments(
            make_segment(-230, 25, 140), make_segment(-240, 85, 145)
        )
        self.assertNotEqual(cascade, compare_to)
        cascade.add_segments(make_segment(-230, 25, 140))
        self.assertNotEqual(cascade, compare_to)
        cascade.add_segments(make_segment(-240, 85, 145))
        self.assertEqual(cascade, compare_to)


if __name__ == "__main__":
    unittest.main()



def make_segment(heat_flow, supply_temp, target_temp, temp_shift=None):
    """
    Creates a new segment with the given properties.
    """
    if supply_temp == target_temp:
        from pina.segments.latent_segment import LatentSegment

        cls = LatentSegment
    else:
        from pina.segments.sensible_segment import SensibleSegment

        cls = SensibleSegment

    return cls.new(heat_flow, supply_temp, target_temp, temp_shift)



from pina.enums import HeatType
from pina.segments.base_segment import BaseSegment


class LatentSegment(BaseSegment):
    """
    Latent segment of a stream in which its temperature does not change.
    """

    def __init__(self, heat_flow, supply_temp, temp_shift=None):
        super().__init__(supply_temp=supply_temp, temp_shift=temp_shift)
        self._heat_flow = heat_flow

    @property
    def heat_type(self):
        return HeatType.LATENT

    @property
    def heat_flow(self):
        return self._heat_flow

    @property
    def target_temp(self):
        # Supply and target temperatures are equal in a latent segment
        return self._supply_temp

    @classmethod
    def new(cls, heat_flow, supply_temp, target_temp, temp_shift=None):
        if supply_temp != target_temp:
            raise ValueError(
                "Temperatures are different: "
                "supply_temp = {}, target_temp = {}".format(supply_temp, target_temp)
            )

        return cls(heat_flow, supply_temp, temp_shift)

    def clone(self):
        return LatentSegment(
            heat_flow=self.heat_flow,
            supply_temp=self.supply_temp,
            temp_shift=self.temp_shift,
        )

    def shift(self, default_temp_shift=None):
        if self.heat_flow == 0:
            shift_by = 0
        elif self.temp_shift is not None:
            shift_by = self.temp_shift
        elif default_temp_shift is not None:
            shift_by = default_temp_shift
        else:
            raise ValueError("No temperature shift given.")

        if self.heat_flow > 0:
            shift_by *= -1

        return LatentSegment(self.heat_flow, self.supply_temp + shift_by, 0)

    def with_low_supply_temp(self):
        return self

    def with_inverted_heat_flow(self):
        return LatentSegment(-self.heat_flow, self.supply_temp, self.temp_shift)

    def split(self, temperatures):
        # There is nothing to split in a latent segment because it has a
        # temperature range of 0, so the returned list always contains the
        # entire segment
        return [self]

    def add(self, other, temp_shift=None):
        if self.heat_type != other.heat_type:
            raise ValueError(
                "Heat type mismatch: {} != {}".format(self.heat_type, other.heat_type)
            )
        elif self.supply_temp != other.supply_temp:
            raise ValueError(
                "Temperature mismatch: {} != {}".format(
                    self.supply_temp, other.supply_temp
                )
            )

        return LatentSegment(
            self.heat_flow + other.heat_flow, self.supply_temp, temp_shift
        )

    def link(self, other, temp_shift=None):
        return self.add(other, temp_shift)

    def __repr__(self):
        return "{}({}, {}, {})".format(
            type(self).__qualname__,
            self._heat_flow,
            self._supply_temp,
            self._temp_shift,
        )



import abc


class BaseSegment(abc.ABC):
    """
    Abstract base class for stream segments
    """

    def __init__(self, supply_temp, temp_shift):
        self._supply_temp = supply_temp
        self._temp_shift = temp_shift

    @property
    @abc.abstractmethod
    def heat_type(self):
        pass

    @property
    @abc.abstractmethod
    def heat_flow(self):
        pass

    @property
    def supply_temp(self):
        return self._supply_temp

    @property
    @abc.abstractmethod
    def target_temp(self):
        pass

    @property
    def temp_shift(self):
        return self._temp_shift

    @property
    def min_temp(self):
        return min(self.supply_temp, self.target_temp)

    @property
    def max_temp(self):
        return max(self.supply_temp, self.target_temp)

    @classmethod
    @abc.abstractmethod
    def new(cls, heat_flow, supply_temp, target_temp, temp_shift):
        pass

    @abc.abstractmethod
    def clone(self):
        pass

    @abc.abstractmethod
    def shift(self, default_temp_shift=None):
        """
        Returns the segment with supply and target temperatures shifted by
        (+ temp_shift) for cold segments and (- temp_shift) for hot segments.
        If the segment's temp_shift is None, the default_temp_shift is used
        instead. If it is also None, a ValueError is raised. The temp_shift of
        the returned segment is set to 0.
        """
        pass

    @abc.abstractmethod
    def with_low_supply_temp(self):
        """
        Returns a segment like self, but with equal supply and minimum
        temperature and equal target and maximum temperature. The supply and
        target temperatures may be swapped compared to self. The heat flow
        remains the same (its sign does not change).
        """
        pass

    def with_absolute_heat_flow(self):
        """
        Returns the segment with absolute heat flow. The supply and target
        temperatures remain the same (they are not swapped).
        """
        if self.heat_flow < 0:
            return self.with_inverted_heat_flow()
        else:
            return self

    @abc.abstractmethod
    def with_inverted_heat_flow(self):
        """
        Returns the segment with inverted heat flow. The supply and target
        temperatures remain the same (they are not swapped).
        """
        pass

    @abc.abstractmethod
    def split(self, temperatures):
        """
        Splits the segment at all given temperatures and returns a list of the
        resulting subsegments. If none of the given temperatures is within the
        temperature range of the segment or temperatures is empty, the returned
        list contains the entire segment as the only element.
        """
        pass

    @abc.abstractmethod
    def add(self, other, temp_shift):
        """
        Returns a segment with the added heat flows of self and other if both
        have equal minimum temperature and equal maximum temperature, otherwise
        returns a ValueError.
        """
        pass

    @abc.abstractmethod
    def link(self, other, temp_shift):
        """
        If self and other have equal heat capacity flow rate and one's supply
        temperature equals the other's target temperature then this method
        returns a segment with the same heat capacity flow rate and the
        differing supply and target temperatures. Otherwise returns a
        ValueError.
        """
        pass

    def __eq__(self, other):
        equal = self.heat_type == other.heat_type
        equal &= self.heat_flow == other.heat_flow
        equal &= self.supply_temp == other.supply_temp
        equal &= self.target_temp == other.target_temp
        equal &= self.temp_shift == other.temp_shift

        return equal



import unittest

from pina.enums import HeatType
from pina.segments.latent_segment import LatentSegment
from pina.segments.sensible_segment import SensibleSegment


class TestLatentSegment(unittest.TestCase):
    """
    Test class for LatentSegment
    """

    def setUp(self):
        self.neutral_segment = LatentSegment(0, 80)
        self.cold_segment = LatentSegment(-200, 100)
        self.hot_segment = LatentSegment(300, 150)

    def tearDown(self):
        pass

    def test_heat_type(self):
        self.assertEqual(self.neutral_segment.heat_type, HeatType.LATENT)
        self.assertEqual(self.cold_segment.heat_type, HeatType.LATENT)
        self.assertEqual(self.hot_segment.heat_type, HeatType.LATENT)

    def test_supply_temp(self):
        self.assertEqual(self.neutral_segment.supply_temp, 80)
        self.assertEqual(self.cold_segment.supply_temp, 100)
        self.assertEqual(self.hot_segment.supply_temp, 150)

    def test_target_temp(self):
        self.assertEqual(self.neutral_segment.target_temp, 80)
        self.assertEqual(self.cold_segment.target_temp, 100)
        self.assertEqual(self.hot_segment.target_temp, 150)

    def test_heat_flow(self):
        self.assertEqual(self.neutral_segment.heat_flow, 0)
        self.assertEqual(self.cold_segment.heat_flow, -200)
        self.assertEqual(self.hot_segment.heat_flow, 300)

    def test_default_temp_shift(self):
        self.assertEqual(self.neutral_segment.temp_shift, None)
        self.assertEqual(self.cold_segment.temp_shift, None)
        self.assertEqual(self.hot_segment.temp_shift, None)

    def test_temp_shift(self):
        segment = LatentSegment(100, 200, 10)
        self.assertEqual(segment.temp_shift, 10)

    def test_min_and_max_temps(self):
        self.assertEqual(self.neutral_segment.min_temp, 80)
        self.assertEqual(self.neutral_segment.max_temp, 80)
        self.assertEqual(self.cold_segment.min_temp, 100)
        self.assertEqual(self.cold_segment.max_temp, 100)
        self.assertEqual(self.hot_segment.min_temp, 150)
        self.assertEqual(self.hot_segment.max_temp, 150)

    def test_new_equal_temps(self):
        self.assertEqual(
            LatentSegment.new(400, 100, 100), LatentSegment(400, 100, None)
        )
        self.assertEqual(LatentSegment.new(-500, 0, 0, 10), LatentSegment(-500, 0, 10))

    def test_new_different_temps(self):
        with self.assertRaises(ValueError):
            LatentSegment.new(400, 100, 150)
        with self.assertRaises(ValueError):
            LatentSegment.new(-500, 0, -50, 10)

    def test_clone(self):
        original = LatentSegment.new(400, 100, 100)
        clone = original.clone()
        self.assertEqual(original, clone)
        self.assertIsNot(original, clone)

        original = LatentSegment.new(-500, 0, 0, 10)
        clone = original.clone()
        self.assertEqual(original, clone)
        self.assertIsNot(original, clone)

    def test_shift_no_temp_shift_given(self):
        self.assertEqual(LatentSegment(0, 80).shift(), LatentSegment(0, 80, 0))
        with self.assertRaises(ValueError):
            LatentSegment(-200, 100).shift()
        with self.assertRaises(ValueError):
            LatentSegment(300, 150).shift()

    def test_shift_self_temp_shift(self):
        self.assertEqual(LatentSegment(0, 80, 10).shift(), LatentSegment(0, 80, 0))
        self.assertEqual(
            LatentSegment(-200, 100, 10).shift(), LatentSegment(-200, 110, 0)
        )
        self.assertEqual(
            LatentSegment(300, 150, 10).shift(), LatentSegment(300, 140, 0)
        )

    def test_shift_default_temp_shift(self):
        self.assertEqual(LatentSegment(0, 80).shift(5), LatentSegment(0, 80, 0))
        self.assertEqual(LatentSegment(-200, 100).shift(5), LatentSegment(-200, 105, 0))
        self.assertEqual(LatentSegment(300, 150).shift(5), LatentSegment(300, 145, 0))

    def test_shift_self_and_default_temp_shift(self):
        self.assertEqual(LatentSegment(0, 80, 10).shift(5), LatentSegment(0, 80, 0))
        self.assertEqual(
            LatentSegment(-200, 100, 10).shift(5), LatentSegment(-200, 110, 0)
        )
        self.assertEqual(
            LatentSegment(300, 150, 10).shift(5), LatentSegment(300, 140, 0)
        )

    def test_split(self):
        self.assertEqual(self.neutral_segment.split([]), [self.neutral_segment])
        self.assertEqual(
            self.neutral_segment.split([50, 80, 90]), [self.neutral_segment]
        )

        segment_temp_included = [100, 150]
        segment_temp_not_included = [90, 140]
        self.assertEqual(self.cold_segment.split([]), [self.cold_segment])
        self.assertEqual(
            self.cold_segment.split(segment_temp_included), [self.cold_segment]
        )
        self.assertEqual(
            self.cold_segment.split(segment_temp_not_included), [self.cold_segment]
        )

        self.assertEqual(self.hot_segment.split([]), [self.hot_segment])
        self.assertEqual(
            self.hot_segment.split(segment_temp_included), [self.hot_segment]
        )
        self.assertEqual(
            self.hot_segment.split(segment_temp_not_included), [self.hot_segment]
        )

    def test_with_low_supply_temp(self):
        self.assertEqual(
            self.neutral_segment.with_low_supply_temp(), self.neutral_segment
        )
        self.assertEqual(self.cold_segment.with_low_supply_temp(), self.cold_segment)
        self.assertEqual(self.hot_segment.with_low_supply_temp(), self.hot_segment)

    def test_with_absolute_heat_flow(self):
        self.assertEqual(
            LatentSegment(0, 80).with_absolute_heat_flow(), LatentSegment(0, 80)
        )
        self.assertEqual(
            LatentSegment(-200, 100).with_absolute_heat_flow(), LatentSegment(200, 100)
        )
        self.assertEqual(
            LatentSegment(-200, 100, 5).with_absolute_heat_flow(),
            LatentSegment(200, 100, 5),
        )
        self.assertEqual(
            LatentSegment(300, 150).with_absolute_heat_flow(), LatentSegment(300, 150)
        )
        self.assertEqual(
            LatentSegment(300, 150, 10).with_absolute_heat_flow(),
            LatentSegment(300, 150, 10),
        )

    def test_with_inverted_heat_flow(self):
        self.assertEqual(
            LatentSegment(0, 80).with_inverted_heat_flow(), LatentSegment(0, 80)
        )
        self.assertEqual(
            LatentSegment(-200, 100).with_inverted_heat_flow(), LatentSegment(200, 100)
        )
        self.assertEqual(
            LatentSegment(-200, 100, 5).with_inverted_heat_flow(),
            LatentSegment(200, 100, 5),
        )
        self.assertEqual(
            LatentSegment(300, 150).with_inverted_heat_flow(), LatentSegment(-300, 150)
        )
        self.assertEqual(
            LatentSegment(300, 150, 10).with_inverted_heat_flow(),
            LatentSegment(-300, 150, 10),
        )

    def test_add(self):
        with self.assertRaises(ValueError):
            self.cold_segment.add(self.neutral_segment)
        with self.assertRaises(ValueError):
            self.neutral_segment.add(self.hot_segment)
        with self.assertRaises(ValueError):
            self.cold_segment.add(self.hot_segment)
        with self.assertRaises(ValueError):
            self.hot_segment.add(self.cold_segment, 10)

        self.assertEqual(
            self.neutral_segment.add(self.neutral_segment), LatentSegment(0, 80)
        )

        self.assertEqual(
            LatentSegment(200, 100).add(LatentSegment(50, 100)), LatentSegment(250, 100)
        )
        self.assertEqual(
            LatentSegment(200, 100, 20).add(LatentSegment(-150, 100, 5), 10),
            LatentSegment(50, 100, 10),
        )

    def test_link(self):
        with self.assertRaises(ValueError):
            self.cold_segment.link(self.neutral_segment)
        with self.assertRaises(ValueError):
            self.neutral_segment.link(self.hot_segment)
        with self.assertRaises(ValueError):
            self.cold_segment.link(self.hot_segment)
        with self.assertRaises(ValueError):
            self.hot_segment.link(self.cold_segment, 10)

        self.assertEqual(
            self.neutral_segment.link(self.neutral_segment), LatentSegment(0, 80)
        )

        self.assertEqual(
            LatentSegment(200, 100).link(LatentSegment(50, 100)),
            LatentSegment(250, 100),
        )
        self.assertEqual(
            LatentSegment(200, 100, 20).link(LatentSegment(-150, 100, 5), 10),
            LatentSegment(50, 100, 10),
        )

    def test_equality_comparison(self):
        self.assertEqual(self.neutral_segment, LatentSegment(0, 80))
        self.assertNotEqual(self.neutral_segment, LatentSegment(0, 80, 10))
        self.assertNotEqual(self.neutral_segment, LatentSegment(100, 80))
        self.assertNotEqual(self.neutral_segment, LatentSegment(0, 120))
        self.assertNotEqual(self.neutral_segment, SensibleSegment(2, 50, 200))

        self.assertEqual(self.cold_segment, LatentSegment(-200, 100))
        self.assertNotEqual(self.cold_segment, LatentSegment(-200, 100, 10))
        self.assertNotEqual(self.cold_segment, LatentSegment(-300, 100))
        self.assertNotEqual(self.cold_segment, LatentSegment(-200, 150))
        self.assertNotEqual(self.cold_segment, SensibleSegment(4, 100, 150))

        self.assertEqual(self.hot_segment, LatentSegment(300, 150))
        self.assertNotEqual(self.hot_segment, LatentSegment(300, 150, 5))
        self.assertNotEqual(self.hot_segment, LatentSegment(200, 150))
        self.assertNotEqual(self.hot_segment, LatentSegment(300, 240))
        self.assertNotEqual(self.cold_segment, SensibleSegment(6, 150, 100))


if __name__ == "__main__":
    unittest.main()



from pina.enums import HeatType
from pina.segments.base_segment import BaseSegment


class SensibleSegment(BaseSegment):
    """
    Sensible segment of a stream in which its temperature changes.
    """

    def __init__(
        self, heat_capacity_flow_rate, supply_temp, target_temp, temp_shift=None
    ):
        super().__init__(supply_temp=supply_temp, temp_shift=temp_shift)
        self._target_temp = target_temp
        self._heat_capacity_flow_rate = heat_capacity_flow_rate

    @property
    def heat_type(self):
        return HeatType.SENSIBLE

    @property
    def heat_capacity_flow_rate(self):
        return self._heat_capacity_flow_rate

    @property
    def heat_flow(self):
        return self._heat_capacity_flow_rate * (self._supply_temp - self._target_temp)

    @property
    def target_temp(self):
        return self._target_temp

    @classmethod
    def new(cls, heat_flow, supply_temp, target_temp, temp_shift=None):
        if supply_temp == target_temp:
            raise ValueError(
                "Temperatures are equal: supply_temp = {}, target_temp = {}".format(
                    supply_temp, target_temp
                )
            )

        heat_capacity_flow_rate = heat_flow / (supply_temp - target_temp)
        return cls(heat_capacity_flow_rate, supply_temp, target_temp, temp_shift)

    def clone(self):
        return SensibleSegment(
            heat_capacity_flow_rate=self.heat_capacity_flow_rate,
            supply_temp=self.supply_temp,
            target_temp=self.target_temp,
            temp_shift=self.temp_shift,
        )

    def shift(self, default_temp_shift=None):
        if self.heat_flow == 0:
            shift_by = 0
        elif self.temp_shift is not None:
            shift_by = self.temp_shift
        elif default_temp_shift is not None:
            shift_by = default_temp_shift
        else:
            raise ValueError("No temperature shift given.")

        if self.heat_flow > 0:
            shift_by *= -1

        return SensibleSegment(
            self.heat_capacity_flow_rate,
            self.supply_temp + shift_by,
            self.target_temp + shift_by,
            0,
        )

    def with_low_supply_temp(self):
        if self.supply_temp > self.target_temp:
            return SensibleSegment(
                -self.heat_capacity_flow_rate,
                self.target_temp,
                self.supply_temp,
                self.temp_shift,
            )
        else:
            return self

    def with_inverted_heat_flow(self):
        return SensibleSegment(
            -self.heat_capacity_flow_rate,
            self.supply_temp,
            self.target_temp,
            self.temp_shift,
        )

    def split(self, temperatures):
        if self.supply_temp < self.target_temp:
            return self._split_ascending(temperatures)
        elif self.supply_temp > self.target_temp:
            return self._split_descending(temperatures)
        else:
            return [self]

    def add(self, other, temp_shift=None):
        if self.heat_type != other.heat_type:
            raise ValueError(
                "Heat type mismatch: {} != {}".format(self.heat_type, other.heat_type)
            )
        elif self.min_temp != other.min_temp:
            raise ValueError(
                "Minimum temperature mismatch: {} != {}".format(
                    self.min_temp, other.min_temp
                )
            )
        elif self.max_temp != other.max_temp:
            raise ValueError(
                "Maximum temperature mismatch: {} != {}".format(
                    self.max_temp, other.max_temp
                )
            )

        if self.supply_temp == other.supply_temp:
            heat_capacity_flow_rate = (
                self.heat_capacity_flow_rate + other.heat_capacity_flow_rate
            )
        else:
            heat_capacity_flow_rate = (
                self.heat_capacity_flow_rate - other.heat_capacity_flow_rate
            )

        return SensibleSegment(
            heat_capacity_flow_rate, self.supply_temp, self.target_temp, temp_shift
        )

    def link(self, other, temp_shift=None):
        if self.heat_type != other.heat_type:
            raise ValueError(
                "Heat type mismatch: {} != {}".format(self.heat_type, other.heat_type)
            )
        elif self.heat_capacity_flow_rate != other.heat_capacity_flow_rate:
            raise ValueError(
                "Heat capacity flowrate mismatch: {} != {}".format(
                    self.heat_capacity_flow_rate, other.heat_capacity_flow_rate
                )
            )

        temp_match = False
        if self.target_temp == other.supply_temp:
            temp_match = True
            supply_temp = self.supply_temp
            target_temp = other.target_temp
        elif self.supply_temp == other.target_temp:
            temp_match = True
            supply_temp = other.supply_temp
            target_temp = self.target_temp

        if temp_match:
            return SensibleSegment(
                self.heat_capacity_flow_rate, supply_temp, target_temp, temp_shift
            )
        else:
            raise ValueError(
                "No matching supply and target temperatures found:\n"
                "\tself.supply_temp: {}, self.target_temp: {}\n"
                "\tother.supply_temp: {}, other.target_temp: {}\n".format(
                    self.supply_temp,
                    self.target_temp,
                    other.supply_temp,
                    other.target_temp,
                )
            )

    def _split_ascending(self, temperatures):
        if not temperatures:
            return [self]

        subsegments = []
        sorted_temps = sorted(temperatures)
        low_temp = self.min_temp
        high_temp = self.max_temp

        for current_temp in sorted_temps:
            if current_temp <= low_temp:
                continue

            if current_temp > high_temp:
                break

            subsegments.append(
                SensibleSegment(
                    self.heat_capacity_flow_rate,
                    low_temp,
                    current_temp,
                    self.temp_shift,
                )
            )

            low_temp = current_temp

        if low_temp < high_temp:
            subsegments.append(
                SensibleSegment(
                    self.heat_capacity_flow_rate, low_temp, high_temp, self.temp_shift
                )
            )

        return subsegments

    def _split_descending(self, temperatures):
        if not temperatures:
            return [self]

        subsegments = []
        sorted_temps = sorted(temperatures, reverse=True)
        low_temp = self.min_temp
        high_temp = self.max_temp

        for current_temp in sorted_temps:
            if current_temp >= high_temp:
                continue

            if current_temp < low_temp:
                break

            subsegments.append(
                SensibleSegment(
                    self.heat_capacity_flow_rate,
                    high_temp,
                    current_temp,
                    self.temp_shift,
                )
            )

            high_temp = current_temp

        if high_temp > low_temp:
            subsegments.append(
                SensibleSegment(
                    self.heat_capacity_flow_rate, high_temp, low_temp, self.temp_shift
                )
            )

        return subsegments

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and self.heat_capacity_flow_rate == other.heat_capacity_flow_rate
        )

    def __repr__(self):
        return "{}({}, {}, {}, {})".format(
            type(self).__qualname__,
            self._heat_capacity_flow_rate,
            self._supply_temp,
            self._target_temp,
            self._temp_shift,
        )



import unittest

from pina.segments import make_segment
from pina.stream import Stream, make_segmented_stream, make_stream


class TestStream(unittest.TestCase):
    """
    Test class for Stream
    """

    def setUp(self):
        self.cold_segments = [
            make_segment(-320, 20, 100),
            make_segment(-400, 100, 100),
            make_segment(-150, 100, 250),
            make_segment(-195, 250, 380),
        ]

        self.hot_segments = [
            make_segment(300, 250, 100),
            make_segment(300, 100, 100),
            make_segment(125, 100, 50),
        ]

        self.mixed_segments = [make_segment(-320, 20, 100), make_segment(320, 100, 20)]

        self.single_sensible_segment = Stream(*self.cold_segments[:1])
        self.single_latent_segment = Stream(*self.hot_segments[1:2])
        self.cold_stream = Stream(*self.cold_segments)
        self.hot_stream = Stream(*self.hot_segments)
        self.neutral_stream = Stream(*self.mixed_segments)

    def tearDown(self):
        pass

    def test_segments_empty(self):
        with self.assertRaises(ValueError):
            Stream()

    def test_temperature_mismatch(self):
        mismatch_first = [make_segment(-150, 50, 200), *self.cold_segments]
        with self.assertRaises(ValueError):
            Stream(*mismatch_first)

        mismatch_middle = self.cold_segments
        mismatch_middle[1] = make_segment(-120, 400, 400)
        with self.assertRaises(ValueError):
            Stream(*mismatch_middle)

        mismatch_last = [*self.cold_segments, make_segment(-150, 50, 200)]
        with self.assertRaises(ValueError):
            Stream(*mismatch_last)

    def test_supply_temp(self):
        self.assertEqual(self.single_sensible_segment.supply_temp, 20)
        self.assertEqual(self.single_latent_segment.supply_temp, 100)
        self.assertEqual(self.cold_stream.supply_temp, 20)
        self.assertEqual(self.hot_stream.supply_temp, 250)
        self.assertEqual(self.neutral_stream.supply_temp, 20)

    def test_target_temp(self):
        self.assertEqual(self.single_sensible_segment.target_temp, 100)
        self.assertEqual(self.single_latent_segment.target_temp, 100)
        self.assertEqual(self.cold_stream.target_temp, 380)
        self.assertEqual(self.hot_stream.target_temp, 50)
        self.assertEqual(self.neutral_stream.target_temp, 20)

    def test_heat_flow(self):
        self.assertEqual(self.single_sensible_segment.heat_flow, -320)
        self.assertEqual(self.single_latent_segment.heat_flow, 300)
        self.assertEqual(self.cold_stream.heat_flow, -1065)
        self.assertEqual(self.hot_stream.heat_flow, 725)
        self.assertEqual(self.neutral_stream.heat_flow, 0)

    def test_segments(self):
        self.assertEqual(self.single_sensible_segment.segments, self.cold_segments[:1])
        self.assertEqual(self.single_latent_segment.segments, self.hot_segments[1:2])
        self.assertEqual(self.cold_stream.segments, self.cold_segments)
        self.assertEqual(self.hot_stream.segments, self.hot_segments)
        self.assertEqual(self.neutral_stream.segments, self.mixed_segments)

    def test_segments_by_type(self):
        test_segments = [
            make_segment(-320, 20, 100),
            make_segment(-400, 100, 100),
            make_segment(0, 100, 100),
            make_segment(400, 100, 0),
            make_segment(300, 0, 0),
            make_segment(0, 0, -20),
        ]

        stream = Stream(*test_segments)

        self.assertEqual(stream.segments, test_segments)
        self.assertEqual(
            stream.neutral_segments,
            [make_segment(0, 100, 100), make_segment(0, 0, -20)],
        )
        self.assertEqual(
            stream.cold_segments,
            [make_segment(-320, 20, 100), make_segment(-400, 100, 100)],
        )
        self.assertEqual(
            stream.hot_segments, [make_segment(400, 100, 0), make_segment(300, 0, 0)]
        )

    def test_make_stream(self):
        from_init = Stream(make_segment(-320, 20, 100, 5))
        from_make = make_stream(-320, 20, 100, 5)
        self.assertEqual(from_init.segments, from_make.segments)

        from_init = Stream(make_segment(-400, 100, 100))
        from_make = make_stream(-400, 100, 100)
        self.assertEqual(from_init.segments, from_make.segments)

    def test_make_segmented_stream(self):
        from_init = Stream(
            make_segment(-320, 20, 100, 5),
            make_segment(-400, 100, 100),
            make_segment(0, 100, 100, 10),
            make_segment(400, 100, 0),
            make_segment(300, 0, 0),
            make_segment(0, 0, -20),
        )

        from_make = make_segmented_stream(
            [-320, 20, 100, 5],
            [-400, 100, 100],
            [0, 100, 100, 10],
            [400, 100, 0],
            [300, 0, 0],
            [0, 0, -20],
        )

        self.assertEqual(from_init.segments, from_make.segments)

    def test_equality_comparison(self):
        single_seg_stream = make_stream(-320, 20, 100, 5)
        multi_seg_stream = make_segmented_stream([-320, 20, 100, 5], [-400, 100, 100])

        self.assertEqual(single_seg_stream, make_stream(-320, 20, 100, 5))
        self.assertNotEqual(single_seg_stream, make_stream(-300, 20, 100, 5))
        self.assertNotEqual(single_seg_stream, make_stream(-320, 10, 100, 5))
        self.assertNotEqual(single_seg_stream, make_stream(-320, 20, 90, 5))
        self.assertNotEqual(single_seg_stream, make_stream(-320, 20, 100, 10))
        self.assertNotEqual(single_seg_stream, make_stream(-320, 20, 100, None))

        self.assertEqual(
            multi_seg_stream,
            make_segmented_stream([-320, 20, 100, 5], [-400, 100, 100]),
        )
        self.assertNotEqual(single_seg_stream, multi_seg_stream)

    def test_repr(self):
        exec("from pina.segments.latent_segment import LatentSegment")
        exec("from pina.segments.sensible_segment import SensibleSegment")

        single_seg_stream = make_stream(-320, 20, 100, 5)
        self.assertEqual(eval(repr(single_seg_stream)), single_seg_stream)

        multi_seg_stream = make_segmented_stream([-320, 20, 100, 5], [-400, 100, 100])
        self.assertEqual(eval(repr(multi_seg_stream)), multi_seg_stream)


if __name__ == "__main__":
    unittest.main()



"""A lightweight pinch analysis package"""

from pina.pinch_analyzer import PinchAnalyzer
from pina.stream import make_segmented_stream, make_stream

__version__ = "0.1.0"

__all__ = ["PinchAnalyzer", "make_segmented_stream", "make_stream"]



from matplotlib import pyplot as plt

from pina import PinchAnalyzer, make_stream

# Arguments: heat flow, supply temperature, target temperature
cold_1 = make_stream(-230, 20, 135)
hot_1 = make_stream(330, 170, 60)
cold_2 = make_stream(-240, 80, 140)
hot_2 = make_stream(180, 150, 30)

min_temp_diff = 10
temp_shift = min_temp_diff / 2

analyzer = PinchAnalyzer(temp_shift)
analyzer.add_streams(cold_1, hot_1, cold_2, hot_2)

print(
    "Heating demand:       {}\n"
    "Cooling demand:       {}\n"
    "Hot utility target:   {}\n"
    "Cold utility target:  {}\n"
    "Heat recovery target: {}\n"
    "Pinch temperature(s): {}".format(
        analyzer.heating_demand,
        analyzer.cooling_demand,
        analyzer.hot_utility_target,
        analyzer.cold_utility_target,
        analyzer.heat_recovery_target,
        analyzer.pinch_temps,
    )
)


fig, ax = plt.subplots(1, 2, figsize=(16, 7))

ax[0].plot(*analyzer.hot_composite_curve, color="tab:red", linestyle="--", label="HCC")
ax[0].plot(*analyzer.cold_composite_curve, color="tab:blue", linestyle="-", label="CCC")
ax[0].legend()
ax[0].set_title("Hot and cold composite curves")
ax[0].set_xlabel("Heat flow [kW]")
ax[0].set_ylabel("Actual temperature [\u2103]")

ax[1].plot(*analyzer.grand_composite_curve, color="k", linestyle="-", label="GCC")
ax[1].legend()
ax[1].set_title("Grand composite curve")
ax[1].set_xlabel("Net heat flow [kW]")
ax[1].set_ylabel("Shifted temperature [\u2103]")

# Make the y-axis equal in both plots
ylims = (*ax[0].get_ylim(), *ax[1].get_ylim())
minmax_ylims = (min(ylims), max(ylims))
ax[0].set_ylim(minmax_ylims)
ax[1].set_ylim(minmax_ylims)

plt.show()



class HeatCascade:
    """
    Heat cascade consisting of temperature intervals, each of which has a
    constant heat capacity flow rate.
    """

    def __init__(self):
        # List of intervals (which are simply segments) in the cascade
        self._intervals = []

    @property
    def intervals(self):
        return self._intervals

    def add_segments(self, *segments):
        """
        Adds segments to the heat cascade, i.e. each segment's heat flow is
        added at the respective temperature range.
        """
        for s in segments:
            self._add_one(s)

    def net_heat_flow(self):
        """
        Returns the net heat flow summed over all intervals.
        """
        return sum(i.heat_flow for i in self.intervals)

    def cumulative_heat_flow(self, heat_offset=0):
        """
        Returns a tuple of two lists:
        * The first list contains the cumulative heat flows at the beginning
          and end of each interval.
        * The second list contains the corresponding temperatures.
        The two lists are sorted by temperature, from lowest to highest.
        The cumulative heat flow at the lowest temperature is set to
        heat_offset.

        The two lists form the coordinates of the composite curves.
        """
        temperatures = []
        heat_flows = []
        if self.intervals:
            temperatures.append(self.intervals[0].supply_temp)
            heat_flows.append(heat_offset)
            for i in self.intervals:
                if i.supply_temp != temperatures[-1]:
                    # There is a temperature gap between 2 intervals
                    temperatures.append(i.supply_temp)
                    heat_flows.append(heat_flows[-1])
                temperatures.append(i.target_temp)
                heat_flows.append(heat_flows[-1] + i.heat_flow)

        return heat_flows, temperatures

    def _add_one(self, segment):
        # Split new segment and existing intervals to get rid of overlaps
        subsegments = segment.with_low_supply_temp().split(
            temperature
            for interval in self.intervals
            for temperature in [interval.min_temp, interval.max_temp]
        )
        self._intervals = [
            new_interval
            for old_interval in self._intervals
            for new_interval in old_interval.split([segment.min_temp, segment.max_temp])
        ]

        self._add_tailored(subsegments)
        self._link_intervals()

    def _add_tailored(self, segments):
        # Add the segments to the heat cascade.
        #
        # The temperature range of each segment and each existing interval must
        # not overlap. They must either match exactly or be separated.
        index = 0
        for s in segments:
            while index < len(self.intervals):
                if s.min_temp >= self.intervals[index].max_temp:
                    # The segment will be added at higher temperatures
                    index += 1
                    continue

                try:
                    # Add segment to interval
                    self._intervals[index] = self.intervals[index].add(s)
                except ValueError:
                    # No matching interval found, insert new one
                    self._intervals.insert(index, s)

                break
            else:
                # No matching interval found, append new one
                self._intervals.append(s)

    def _link_intervals(self):
        # Links adjacent intervals, if possible.
        index = 0
        while index < len(self.intervals) - 1:
            try:
                self._intervals[index] = self.intervals[index].link(
                    self.intervals[index + 1]
                )
            except ValueError:
                pass
            else:
                self._intervals.pop(index + 1)

            if self.intervals[index].heat_flow == 0:
                self._intervals.pop(index)
                continue

            index += 1

        if self.intervals and self.intervals[-1].heat_flow == 0:
            self._intervals.pop(-1)

    def __eq__(self, other):
        return self.intervals == other.intervals



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pina', 'pina.segments']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

extras_require = \
{'dev': ['black', 'coverage', 'flake8', 'isort', 'pre-commit', 'tox']}

setup(name='pina',
      version='0.1.0',
      description='A lightweight pinch analysis package',
      author='Patrick Heck',
      author_email='fitz_oreilly@outlook.com',
      url=None,
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



import unittest

from pina.enums import HeatType
from pina.segments.latent_segment import LatentSegment
from pina.segments.sensible_segment import SensibleSegment


class TestSensibleSegment(unittest.TestCase):
    """
    Test class for SensibleSegment
    """

    def setUp(self):
        self.neutral_segment = SensibleSegment(2, 50, 50)
        self.zero_capacity_segment = SensibleSegment(0, 80, 120)
        self.cold_segment = SensibleSegment(1, 20, 200)
        self.hot_segment = SensibleSegment(1.8, 150, 50)

    def tearDown(self):
        pass

    def test_heat_type(self):
        self.assertEqual(self.neutral_segment.heat_type, HeatType.SENSIBLE)
        self.assertEqual(self.zero_capacity_segment.heat_type, HeatType.SENSIBLE)
        self.assertEqual(self.cold_segment.heat_type, HeatType.SENSIBLE)
        self.assertEqual(self.hot_segment.heat_type, HeatType.SENSIBLE)

    def test_heat_capacity_flow_rate(self):
        self.assertEqual(self.neutral_segment.heat_capacity_flow_rate, 2)
        self.assertEqual(self.zero_capacity_segment.heat_capacity_flow_rate, 0)
        self.assertEqual(self.cold_segment.heat_capacity_flow_rate, 1)
        self.assertEqual(self.hot_segment.heat_capacity_flow_rate, 1.8)

    def test_supply_temp(self):
        self.assertEqual(self.neutral_segment.supply_temp, 50)
        self.assertEqual(self.zero_capacity_segment.supply_temp, 80)
        self.assertEqual(self.cold_segment.supply_temp, 20)
        self.assertEqual(self.hot_segment.supply_temp, 150)

    def test_target_temp(self):
        self.assertEqual(self.neutral_segment.target_temp, 50)
        self.assertEqual(self.zero_capacity_segment.target_temp, 120)
        self.assertEqual(self.cold_segment.target_temp, 200)
        self.assertEqual(self.hot_segment.target_temp, 50)

    def test_heat_flow(self):
        self.assertEqual(self.neutral_segment.heat_flow, 0)
        self.assertEqual(self.zero_capacity_segment.heat_flow, 0)
        self.assertEqual(self.cold_segment.heat_flow, -180)
        self.assertEqual(self.hot_segment.heat_flow, 180)

    def test_default_temp_shift(self):
        self.assertEqual(self.cold_segment.temp_shift, None)
        self.assertEqual(self.hot_segment.temp_shift, None)

    def test_temp_shift(self):
        segment = SensibleSegment(1, 20, 200, 5)
        self.assertEqual(segment.temp_shift, 5)

    def test_min_temp(self):
        self.assertEqual(self.neutral_segment.min_temp, 50)
        self.assertEqual(self.zero_capacity_segment.min_temp, 80)
        self.assertEqual(self.cold_segment.min_temp, 20)
        self.assertEqual(self.hot_segment.min_temp, 50)

    def test_max_temp(self):
        self.assertEqual(self.neutral_segment.max_temp, 50)
        self.assertEqual(self.zero_capacity_segment.max_temp, 120)
        self.assertEqual(self.cold_segment.max_temp, 200)
        self.assertEqual(self.hot_segment.max_temp, 150)

    def test_new_equal_temps(self):
        with self.assertRaises(ValueError):
            SensibleSegment.new(400, 100, 100)
        with self.assertRaises(ValueError):
            SensibleSegment.new(-500, -50, -50, 10)

    def test_new_different_temps(self):
        self.assertEqual(
            SensibleSegment.new(-400, 100, 200), SensibleSegment(4, 100, 200, None)
        )
        self.assertEqual(
            SensibleSegment.new(500, 100, 0, 10), SensibleSegment(5, 100, 0, 10)
        )

    def test_clone(self):
        original = SensibleSegment.new(400, 100, 200)
        clone = original.clone()
        self.assertEqual(original, clone)
        self.assertIsNot(original, clone)

        original = SensibleSegment.new(-500, 100, 0, 10)
        clone = original.clone()
        self.assertEqual(original, clone)
        self.assertIsNot(original, clone)

    def test_shift_no_temp_shift_given(self):
        self.assertEqual(
            SensibleSegment(2, 50, 50).shift(), SensibleSegment(2, 50, 50, 0)
        )
        self.assertEqual(
            SensibleSegment(0, 80, 120).shift(), SensibleSegment(0, 80, 120, 0)
        )
        with self.assertRaises(ValueError):
            SensibleSegment(1, 20, 200).shift()
        with self.assertRaises(ValueError):
            SensibleSegment(1.8, 150, 50).shift()

    def test_shift_self_temp_shift(self):
        self.assertEqual(
            SensibleSegment(2, 50, 50, 10).shift(), SensibleSegment(2, 50, 50, 0)
        )
        self.assertEqual(
            SensibleSegment(0, 80, 120, 10).shift(), SensibleSegment(0, 80, 120, 0)
        )
        self.assertEqual(
            SensibleSegment(1, 20, 200, 10).shift(), SensibleSegment(1, 30, 210, 0)
        )
        self.assertEqual(
            SensibleSegment(1.8, 150, 50, 10).shift(), SensibleSegment(1.8, 140, 40, 0)
        )

    def test_shift_default_temp_shift(self):
        self.assertEqual(
            SensibleSegment(2, 50, 50).shift(5), SensibleSegment(2, 50, 50, 0)
        )
        self.assertEqual(
            SensibleSegment(0, 80, 120).shift(5), SensibleSegment(0, 80, 120, 0)
        )
        self.assertEqual(
            SensibleSegment(1, 20, 200).shift(5), SensibleSegment(1, 25, 205, 0)
        )
        self.assertEqual(
            SensibleSegment(1.8, 150, 50).shift(5), SensibleSegment(1.8, 145, 45, 0)
        )

    def test_shift_self_and_default_temp_shift(self):
        self.assertEqual(
            SensibleSegment(2, 50, 50, 10).shift(5), SensibleSegment(2, 50, 50, 0)
        )
        self.assertEqual(
            SensibleSegment(0, 80, 120, 10).shift(5), SensibleSegment(0, 80, 120, 0)
        )
        self.assertEqual(
            SensibleSegment(1, 20, 200, 10).shift(5), SensibleSegment(1, 30, 210, 0)
        )
        self.assertEqual(
            SensibleSegment(1.8, 150, 50, 10).shift(5), SensibleSegment(1.8, 140, 40, 0)
        )

    def test_split(self):
        # Neutral segment
        self.assertEqual(self.neutral_segment.split([]), [self.neutral_segment])
        self.assertEqual(
            self.neutral_segment.split([10, 50, 100]), [self.neutral_segment]
        )

        # Neutral segment with zero heat capacity flow rate
        self.assertEqual(
            self.zero_capacity_segment.split([]), [self.zero_capacity_segment]
        )
        self.assertEqual(
            self.zero_capacity_segment.split([10, 50, 100]),
            [SensibleSegment(0, 80, 100), SensibleSegment(0, 100, 120)],
        )

        # Cold segment
        heat_capacity_flow_rate = self.cold_segment.heat_capacity_flow_rate
        min_temp = self.cold_segment.min_temp
        max_temp = self.cold_segment.max_temp

        match_min_max = [20, 200]
        beyond_min_max = [-25, 430]
        below_min = [-10, 10]
        above_max = [250, 400, 600]
        include_min = [-10, 50]
        include_max = [150, 400, 600]
        within_min_max = [80, 120, 150]
        duplicates = [80, 120, 120]
        unsorted = [150, 80, 120]

        self.assertEqual(self.cold_segment.split([]), [self.cold_segment])
        self.assertEqual(self.cold_segment.split(match_min_max), [self.cold_segment])
        self.assertEqual(self.cold_segment.split(beyond_min_max), [self.cold_segment])
        self.assertEqual(self.cold_segment.split(below_min), [self.cold_segment])
        self.assertEqual(self.cold_segment.split(above_max), [self.cold_segment])

        self.assertEqual(
            self.cold_segment.split(include_min),
            [
                SensibleSegment(heat_capacity_flow_rate, min_temp, 50),
                SensibleSegment(heat_capacity_flow_rate, 50, max_temp),
            ],
        )

        self.assertEqual(
            self.cold_segment.split(include_max),
            [
                SensibleSegment(heat_capacity_flow_rate, min_temp, 150),
                SensibleSegment(heat_capacity_flow_rate, 150, max_temp),
            ],
        )

        self.assertEqual(
            self.cold_segment.split(within_min_max),
            [
                SensibleSegment(heat_capacity_flow_rate, min_temp, 80),
                SensibleSegment(heat_capacity_flow_rate, 80, 120),
                SensibleSegment(heat_capacity_flow_rate, 120, 150),
                SensibleSegment(heat_capacity_flow_rate, 150, max_temp),
            ],
        )

        self.assertEqual(
            self.cold_segment.split(duplicates),
            [
                SensibleSegment(heat_capacity_flow_rate, min_temp, 80),
                SensibleSegment(heat_capacity_flow_rate, 80, 120),
                SensibleSegment(heat_capacity_flow_rate, 120, max_temp),
            ],
        )

        self.assertEqual(
            self.cold_segment.split(unsorted),
            [
                SensibleSegment(heat_capacity_flow_rate, min_temp, 80),
                SensibleSegment(heat_capacity_flow_rate, 80, 120),
                SensibleSegment(heat_capacity_flow_rate, 120, 150),
                SensibleSegment(heat_capacity_flow_rate, 150, max_temp),
            ],
        )

        # Hot segment
        heat_capacity_flow_rate = self.hot_segment.heat_capacity_flow_rate
        min_temp = self.hot_segment.min_temp
        max_temp = self.hot_segment.max_temp

        match_min_max = [150, 50]
        beyond_min_max = [-25, 430]
        below_min = [-10, 10]
        above_max = [250, 400, 600]
        include_min = [-10, 80]
        include_max = [120, 400, 600]
        within_min_max = [80, 120, 135]
        duplicates = [80, 120, 120]
        unsorted = [150, 80, 120]

        self.assertEqual(self.hot_segment.split([]), [self.hot_segment])
        self.assertEqual(self.hot_segment.split(match_min_max), [self.hot_segment])
        self.assertEqual(self.hot_segment.split(beyond_min_max), [self.hot_segment])
        self.assertEqual(self.hot_segment.split(below_min), [self.hot_segment])
        self.assertEqual(self.hot_segment.split(above_max), [self.hot_segment])

        self.assertEqual(
            self.hot_segment.split(include_min),
            [
                SensibleSegment(heat_capacity_flow_rate, max_temp, 80),
                SensibleSegment(heat_capacity_flow_rate, 80, min_temp),
            ],
        )

        self.assertEqual(
            self.hot_segment.split(include_max),
            [
                SensibleSegment(heat_capacity_flow_rate, max_temp, 120),
                SensibleSegment(heat_capacity_flow_rate, 120, min_temp),
            ],
        )

        self.assertEqual(
            self.hot_segment.split(within_min_max),
            [
                SensibleSegment(heat_capacity_flow_rate, max_temp, 135),
                SensibleSegment(heat_capacity_flow_rate, 135, 120),
                SensibleSegment(heat_capacity_flow_rate, 120, 80),
                SensibleSegment(heat_capacity_flow_rate, 80, min_temp),
            ],
        )

        self.assertEqual(
            self.hot_segment.split(duplicates),
            [
                SensibleSegment(heat_capacity_flow_rate, max_temp, 120),
                SensibleSegment(heat_capacity_flow_rate, 120, 80),
                SensibleSegment(heat_capacity_flow_rate, 80, min_temp),
            ],
        )

        self.assertEqual(
            self.hot_segment.split(unsorted),
            [
                SensibleSegment(heat_capacity_flow_rate, max_temp, 120),
                SensibleSegment(heat_capacity_flow_rate, 120, 80),
                SensibleSegment(heat_capacity_flow_rate, 80, min_temp),
            ],
        )

    def test_with_low_supply_temp(self):
        self.assertEqual(
            self.neutral_segment.with_low_supply_temp(), self.neutral_segment
        )
        self.assertEqual(self.cold_segment.with_low_supply_temp(), self.cold_segment)
        self.assertEqual(
            self.hot_segment.with_low_supply_temp(), SensibleSegment(-1.8, 50, 150)
        )

    def test_with_absolute_heat_flow(self):
        self.assertEqual(
            SensibleSegment(2, 50, 50).with_absolute_heat_flow(),
            SensibleSegment(2, 50, 50),
        )
        self.assertEqual(
            SensibleSegment(0, 80, 120).with_absolute_heat_flow(),
            SensibleSegment(0, 80, 120),
        )
        self.assertEqual(
            SensibleSegment(1, 20, 200).with_absolute_heat_flow(),
            SensibleSegment(-1, 20, 200),
        )
        self.assertEqual(
            SensibleSegment(-1, 20, 200, 5).with_absolute_heat_flow(),
            SensibleSegment(-1, 20, 200, 5),
        )
        self.assertEqual(
            SensibleSegment(1.8, 150, 50).with_absolute_heat_flow(),
            SensibleSegment(1.8, 150, 50),
        )
        self.assertEqual(
            SensibleSegment(-1.8, 150, 50).with_absolute_heat_flow(),
            SensibleSegment(1.8, 150, 50),
        )
        self.assertEqual(
            SensibleSegment(1.8, 150, 50, 10).with_absolute_heat_flow(),
            SensibleSegment(1.8, 150, 50, 10),
        )

    def test_with_inverted_heat_flow(self):
        self.assertEqual(
            SensibleSegment(2, 50, 50).with_inverted_heat_flow(),
            SensibleSegment(-2, 50, 50),
        )
        self.assertEqual(
            SensibleSegment(0, 80, 120).with_inverted_heat_flow(),
            SensibleSegment(0, 80, 120),
        )
        self.assertEqual(
            SensibleSegment(1, 20, 200).with_inverted_heat_flow(),
            SensibleSegment(-1, 20, 200),
        )
        self.assertEqual(
            SensibleSegment(-1, 20, 200, 5).with_inverted_heat_flow(),
            SensibleSegment(1, 20, 200, 5),
        )
        self.assertEqual(
            SensibleSegment(1.8, 150, 50).with_inverted_heat_flow(),
            SensibleSegment(-1.8, 150, 50),
        )
        self.assertEqual(
            SensibleSegment(-1.8, 150, 50).with_inverted_heat_flow(),
            SensibleSegment(1.8, 150, 50),
        )
        self.assertEqual(
            SensibleSegment(1.8, 150, 50, 10).with_inverted_heat_flow(),
            SensibleSegment(-1.8, 150, 50, 10),
        )

    def test_add(self):
        with self.assertRaises(ValueError):
            self.cold_segment.add(self.neutral_segment)
        with self.assertRaises(ValueError):
            self.neutral_segment.add(self.hot_segment)
        with self.assertRaises(ValueError):
            self.cold_segment.add(self.hot_segment)
        with self.assertRaises(ValueError):
            self.hot_segment.add(self.cold_segment, 10)

        self.assertEqual(
            self.neutral_segment.add(self.neutral_segment), SensibleSegment(4, 50, 50)
        )
        self.assertEqual(
            SensibleSegment(2, 100, 180).add(SensibleSegment(0, 100, 180)),
            SensibleSegment(2, 100, 180),
        )
        self.assertEqual(
            SensibleSegment(0, 100, 180).add(SensibleSegment(2, 100, 180)),
            SensibleSegment(2, 100, 180),
        )
        self.assertEqual(
            SensibleSegment(2, 100, 180).add(SensibleSegment(5, 100, 180)),
            SensibleSegment(7, 100, 180),
        )
        self.assertEqual(
            SensibleSegment(4, 100, 180, 2).add(SensibleSegment(2.5, 180, 100, 5)),
            SensibleSegment(1.5, 100, 180),
        )
        self.assertEqual(
            SensibleSegment(4, 100, 180, 2).add(SensibleSegment(2.5, 180, 100, 5), 10),
            SensibleSegment(1.5, 100, 180, 10),
        )

    def test_link(self):
        with self.assertRaises(ValueError):
            self.cold_segment.link(self.neutral_segment)
        with self.assertRaises(ValueError):
            self.cold_segment.link(self.hot_segment)
        with self.assertRaises(ValueError):
            self.hot_segment.link(self.cold_segment, 10)
        with self.assertRaises(ValueError):
            SensibleSegment(2, 100, 180).link(SensibleSegment(5, 100, 180))
        with self.assertRaises(ValueError):
            SensibleSegment(2, 100, 180).link(SensibleSegment(5, 150, 220, 5))
        with self.assertRaises(ValueError):
            SensibleSegment(2, 100, 180).link(SensibleSegment(5, 180, 240))
        with self.assertRaises(ValueError):
            SensibleSegment(2, 100, 180).link(SensibleSegment(-2, 180, 220))
        with self.assertRaises(ValueError):
            SensibleSegment(2, 100, 180).link(SensibleSegment(2, 100, 180))
        with self.assertRaises(ValueError):
            SensibleSegment(2, 100, 180).link(SensibleSegment(2, 150, 220))

        self.assertEqual(
            self.neutral_segment.link(self.neutral_segment), SensibleSegment(2, 50, 50)
        )
        self.assertEqual(
            SensibleSegment(0, 80, 180).link(SensibleSegment(0, 180, 120)),
            SensibleSegment(0, 80, 120),
        )
        self.assertEqual(
            SensibleSegment(2, 100, 180).link(SensibleSegment(2, 180, 220)),
            SensibleSegment(2, 100, 220),
        )
        self.assertEqual(
            SensibleSegment(2, 100, 180).link(SensibleSegment(2, 180, 140), 5),
            SensibleSegment(2, 100, 140, 5),
        )
        self.assertEqual(
            SensibleSegment(2, 100, 180).link(SensibleSegment(2, 180, 100)),
            SensibleSegment(2, 100, 100),
        )
        self.assertEqual(
            SensibleSegment(2, 180, 220).link(SensibleSegment(2, 100, 180)),
            SensibleSegment(2, 100, 220),
        )
        self.assertEqual(
            SensibleSegment(2, 180, 140).link(SensibleSegment(2, 100, 180), 5),
            SensibleSegment(2, 100, 140, 5),
        )

    def test_equality_comparison(self):
        self.assertEqual(self.neutral_segment, SensibleSegment(2, 50, 50))
        self.assertNotEqual(self.neutral_segment, SensibleSegment(2, 50, 50, 5))
        self.assertNotEqual(self.neutral_segment, SensibleSegment(3.5, 50, 50))
        self.assertNotEqual(self.neutral_segment, SensibleSegment(2, 40, 50))
        self.assertNotEqual(self.neutral_segment, SensibleSegment(2, 50, 40))
        self.assertNotEqual(self.neutral_segment, SensibleSegment(2, 40, 40))
        self.assertNotEqual(self.neutral_segment, LatentSegment(0, 50))

        self.assertEqual(self.zero_capacity_segment, SensibleSegment(0, 80, 120))
        self.assertNotEqual(self.zero_capacity_segment, SensibleSegment(0, 80, 120, 5))
        self.assertNotEqual(self.zero_capacity_segment, SensibleSegment(3.5, 80, 120))
        self.assertNotEqual(self.zero_capacity_segment, SensibleSegment(0, 40, 120))
        self.assertNotEqual(self.zero_capacity_segment, SensibleSegment(0, 80, 40))
        self.assertNotEqual(self.zero_capacity_segment, LatentSegment(0, 80))
        self.assertNotEqual(self.zero_capacity_segment, LatentSegment(0, 120))

        self.assertEqual(self.cold_segment, SensibleSegment(1, 20, 200))
        self.assertNotEqual(self.cold_segment, SensibleSegment(1, 20, 200, 8))
        self.assertNotEqual(self.cold_segment, SensibleSegment(2, 20, 200))
        self.assertNotEqual(self.cold_segment, SensibleSegment(1, 40, 200))
        self.assertNotEqual(self.cold_segment, SensibleSegment(1, 20, 150))
        self.assertNotEqual(self.cold_segment, LatentSegment(20, 180))
        self.assertNotEqual(self.cold_segment, LatentSegment(200, 180))

        self.assertEqual(self.hot_segment, SensibleSegment(1.8, 150, 50))
        self.assertNotEqual(self.hot_segment, SensibleSegment(1.8, 150, 50, 15))
        self.assertNotEqual(self.hot_segment, SensibleSegment(4, 150, 50))
        self.assertNotEqual(self.hot_segment, SensibleSegment(1.8, 240, 50))
        self.assertNotEqual(self.hot_segment, SensibleSegment(1.8, 150, 80))
        self.assertNotEqual(self.hot_segment, LatentSegment(-180, 150))
        self.assertNotEqual(self.hot_segment, LatentSegment(-180, 50))


if __name__ == "__main__":
    unittest.main()



import unittest

from pina.segments import make_segment
from pina.segments.latent_segment import LatentSegment
from pina.segments.sensible_segment import SensibleSegment


class TestSegmentFactory(unittest.TestCase):
    """
    Test class for segment factory
    """

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_new_latent_segment(self):
        self.assertEqual(make_segment(-200, 100, 100), LatentSegment(-200, 100, None))
        self.assertEqual(make_segment(150, 80, 80), LatentSegment(150, 80, None))

    def test_new_latent_segment_with_temp_shift(self):
        self.assertEqual(make_segment(-200, 100, 100, 10), LatentSegment(-200, 100, 10))
        self.assertEqual(make_segment(150, 80, 80, 10), LatentSegment(150, 80, 10))

    def test_new_sensible_segment(self):
        self.assertEqual(make_segment(-150, 20, 80), SensibleSegment(2.5, 20, 80, None))
        self.assertEqual(make_segment(150, 100, 50), SensibleSegment(3, 100, 50, None))

    def test_new_sensible_segment_with_temp_shift(self):
        self.assertEqual(make_segment(-150, 20, 80, 5), SensibleSegment(2.5, 20, 80, 5))
        self.assertEqual(make_segment(150, 100, 50, 5), SensibleSegment(3, 100, 50, 5))


if __name__ == "__main__":
    unittest.main()
